from __future__ import annotations

from collections.abc import Generator, Mapping
from contextlib import contextmanager
from logging import (
    DEBUG,
    Formatter,
    Logger,
    LoggerAdapter,
    NullHandler,
    StreamHandler,
    getLogger,
)
from sys import stdout
from time import time
from types import TracebackType
from typing import Any

__author__ = "davidricodias"
__copyright__ = "davidricodias"
__license__ = "MIT"


package_logger = getLogger("lobmp")
package_logger.addHandler(
    NullHandler()
)  # By default start the logger pointing to the null handler, in essence doing nothing


def activate_logger() -> None:
    logFormatter = Formatter("%(asctime)s %(name)s [%(threadName)s] [%(levelname)s]  %(message)s")
    stream_handler = StreamHandler(stdout)  # Outputs to standard terminal
    stream_handler.setFormatter(logFormatter)
    if stream_handler not in package_logger.handlers:
        package_logger.removeHandler(NullHandler())
        package_logger.addHandler(stream_handler)


def get_logger() -> Logger:
    return package_logger


def set_logger_level(level: int | str) -> None:
    package_logger.setLevel(level)


# Custom logging class to enable both direct and context manager logging
class ContextLogger(LoggerAdapter):
    @contextmanager
    def timeit(self: ContextLogger, msg: str, level: int | str = DEBUG) -> Generator[Any, Any, Any]:
        self.logger.log(level, f"Started with timing: {msg}")
        try:
            start_time = time()
            yield
            elapsed_time = time() - start_time
            self.logger.log(level, f"Completed with timing: {msg} (Elapsed: {elapsed_time:.2f}s)")
        except Exception as e:
            self.logger.log(level, f"Error in {msg}: {e}", exc_info=True)
            raise e

    def debug(
        self: ContextLogger,
        msg: object,
        *args: object,
        exc_info: bool
        | tuple[type[BaseException], BaseException, TracebackType | None]
        | tuple[None, None, None]
        | BaseException
        | None = None,
        stack_info: bool = False,
        stacklevel: int = 1,
        extra: Mapping[str, object] | None = None,
        **kwargs: object,
    ) -> None:
        self.logger.debug(
            msg,
            *args,
            exc_info=exc_info,
            stack_info=stack_info,
            stacklevel=stacklevel,
            extra=extra,
            **kwargs,
        )

    def info(
        self: ContextLogger,
        msg: object,
        *args: object,
        exc_info: bool
        | tuple[type[BaseException], BaseException, TracebackType | None]
        | tuple[None, None, None]
        | BaseException
        | None = None,
        stack_info: bool = False,
        stacklevel: int = 1,
        extra: Mapping[str, object] | None = None,
        **kwargs: object,
    ) -> None:
        self.logger.info(
            msg,
            *args,
            exc_info=exc_info,
            stack_info=stack_info,
            stacklevel=stacklevel,
            extra=extra,
            **kwargs,
        )

    def warning(
        self: ContextLogger,
        msg: object,
        *args: object,
        exc_info: bool
        | tuple[type[BaseException], BaseException, TracebackType | None]
        | tuple[None, None, None]
        | BaseException
        | None = None,
        stack_info: bool = False,
        stacklevel: int = 1,
        extra: Mapping[str, object] | None = None,
        **kwargs: object,
    ) -> None:
        self.logger.warning(
            msg,
            *args,
            exc_info=exc_info,
            stack_info=stack_info,
            stacklevel=stacklevel,
            extra=extra,
            **kwargs,
        )

    def error(
        self: ContextLogger,
        msg: object,
        *args: object,
        exc_info: bool
        | tuple[type[BaseException], BaseException, TracebackType | None]
        | tuple[None, None, None]
        | BaseException
        | None = None,
        stack_info: bool = False,
        stacklevel: int = 1,
        extra: Mapping[str, object] | None = None,
        **kwargs: object,
    ) -> None:
        self.logger.error(
            msg,
            *args,
            exc_info=exc_info,
            stack_info=stack_info,
            stacklevel=stacklevel,
            extra=extra,
            **kwargs,
        )

    def critical(
        self: ContextLogger,
        msg: object,
        *args: object,
        exc_info: bool
        | tuple[type[BaseException], BaseException, TracebackType | None]
        | tuple[None, None, None]
        | BaseException
        | None = None,
        stack_info: bool = False,
        stacklevel: int = 1,
        extra: Mapping[str, object] | None = None,
        **kwargs: object,
    ) -> None:
        self.logger.critical(
            msg,
            *args,
            exc_info=exc_info,
            stack_info=stack_info,
            stacklevel=stacklevel,
            extra=extra,
            **kwargs,
        )


log = ContextLogger(get_logger(), {})

__all__ = ["log"]



"""
Command Line Interface Module
=================

This module provides functionality to let users call the message processor using a command
line interface.

The `cli` module implements an `argparse` with the follwing flags:
- `--filepath <path>`: path of the file to be processed.
- `--provider <str>`: provider of the data to be processed. Currently *only supports lseg*.

Example:
```
lobmp --filepath download_1.csv
```

Author: davidricodias
"""

from argparse import ArgumentParser
from sys import argv, exit, stderr

from lobmp.main import main

__author__ = "davidricodias"
__copyright__ = "davidricodias"
__license__ = "MIT"


def cli() -> int:
    parser = ArgumentParser(description="Limit Order Book Messages Processor")
    parser.add_argument("filepath", help="Path to the input file.", type=str)
    parser.add_argument("targetdir", help="Directory for the processed data.", type=str)
    parser.add_argument(
        "--target",
        default="parquet",
        choices=["parquet"],
        help="Target format for the processing.",
        type=str,
    )
    parser.add_argument(
        "--provider",
        default="lseg",
        choices=["lseg"],
        help="Provider of the data to be processed. Currently *only supports lseg*.",
        type=str,
    )
    parser.add_argument(
        "--verbose",
        default="NOTSET",
        choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL", "NOTSET"],
        help="Level of logging desired. Default is NOTSET, which is no logging.",
        type=str,
    )

    # Show help if no arguments are provided
    if len(argv) == 1:
        parser.print_help(stderr)
        exit(1)

    args = parser.parse_args()
    return main(args.filepath, args.targetdir, args.verbose)


if __name__ == "__main__":
    cli()



known_fids = {
    "PROD_PERM": ["1", False],
    "DSPLY_NAME": ["3", False],
    "CURRENCY": ["15", True],
    "ACTIV_DATE": ["17", False],
    "RECORDTYPE": ["259", False],
    "RDN_EXCHD2": ["1709", True],
    "QUOTE_DATE": ["3386", False],
    "PROV_SYMB": ["3422", False],
    "PR_RNK_RUL": ["3423", True],
    "OR_RNK_RUL": ["3425", True],
    "MNEMONIC": ["3694", False],
    "QUOTIM_MS": ["3855", False],
    "MKT_STATUS": ["3915", False],
    "TIMACT_MS": ["4148", False],
    "CONTEXT_ID": ["5357", False],
    "DDS_DSO_ID": ["6401", False],
    "SPS_SP_RIC": ["6480", False],
    "BOOK_STATE": ["6516", True],
    "HALT_REASN": ["6517", False],
    "MKT_OR_RUL": ["6519", True],
    "TRD_STATUS": ["6614", True],
    "HALT_RSN": ["6615", True],
    "INST_PHASE": ["8927", True],
    "TIMACT_NS": ["14269", False],
    "BID_TIME": ["266", False],
    "ORDER_PRC": ["3427", False],
    "ORDER_SIDE": ["3428", True],
    "NO_ORD": ["3430", False],
    "ACC_SIZE": ["4356", False],
    "LV_TIM_MS": ["6527", False],
    "LV_DATE": ["6529", False],
    "LV_TIM_NS": ["14268", False],
    "ASK_TIME": ["267", False],
    "ORDBK_VOL": ["3875", False],
}

columns_order = [
    "TICKER",
    "TIMESTAMP",
    "GMT_OFFSET",
    "MARKET_MESSAGE_TYPE",
    "PROD_PERM",
    "DSPLY_NAME",
    "CURRENCY",
    "ACTIV_DATE",
    "RECORDTYPE",
    "RDN_EXCHD2",
    "QUOTE_DATE",
    "PROV_SYMB",
    "PR_RNK_RUL",
    "OR_RNK_RUL",
    "MNEMONIC",
    "QUOTIM_MS",
    "MKT_STATUS",
    "TIMACT_MS",
    "CONTEXT_ID",
    "DDS_DSO_ID",
    "SPS_SP_RIC",
    "BOOK_STATE",
    "HALT_REASN",
    "MKT_OR_RUL",
    "TRD_STATUS",
    "HALT_RSN",
    "INST_PHASE",
    "TIMACT_NS",
    "MAP_ENTRY_TYPE",
    "MAP_ENTRY_KEY",
    "BID_TIME",
    "ASK_TIME",
    "ORDER_PRC",
    "ORDER_SIDE",
    "NO_ORD",
    "ACC_SIZE",
    "LV_TIM_MS",
    "LV_DATE",
    "LV_TIM_NS",
    "ORDBK_VOL",
]

supplement = [
    "TICKER",
    "TIMESTAMP",
    "GMT_OFFSET",
    "MARKET_MESSAGE_TYPE",
    "MAP_ENTRY_TYPE",
    "MAP_ENTRY_KEY",
]



from logging import NOTSET, _levelToName
from pathlib import Path

from lobmp import run
from lobmp.logger import activate_logger, log, set_logger_level

__author__ = "davidricodias"
__copyright__ = "davidricodias"
__license__ = "MIT"


def main(filepath: str, targetdir: str, verbose: int | str = "NOTSET") -> int:
    status: int = 0
    if verbose != _levelToName[NOTSET]:
        activate_logger()
        set_logger_level(verbose)

    input_file_path = Path(filepath)
    output_file_path = (Path(targetdir) / input_file_path.stem).with_suffix(".parquet")

    with log.timeit("Execute run"):
        status = run(input_file_path, output_file_path)
    return status



from lobmp._lobmp import find_market_by_price_lines, flatten_map_entry, flatten_market_by_price, run
from lobmp._version import VERSION

__version__ = VERSION


__all__ = ["find_market_by_price_lines", "flatten_map_entry", "flatten_market_by_price", "run"]



from pathlib import Path

from polars import DataFrame

def find_market_by_price_lines(path: Path) -> list: ...
def flatten_map_entry(map_entry: str) -> list[list]: ...
def flatten_market_by_price(market_by_price: str) -> DataFrame: ...
def run(input_file: Path, output_directory: Path) -> bool: ...



import tomllib
from pathlib import Path

from lobmp._version import VERSION


def test_versions_match():
    cargo = Path().absolute() / "Cargo.toml"
    with open(cargo, "rb") as f:
        data = tomllib.load(f)
        cargo_version = data["package"]["version"]

    assert VERSION == cargo_version



"""lobmp tests

Plese remember to follow the basic four steps:
1. Arrange
2. Act
3. Assert
4. Cleanup
"""

from pathlib import Path

import polars as pl
import pytest
from polars.testing import assert_frame_equal

from lobmp import find_market_by_price_lines, flatten_map_entry, flatten_market_by_price, run
from lobmp.definitions.fids import known_fids, supplement
from lobmp.logger import activate_logger, set_logger_level

activate_logger()
set_logger_level("DEBUG")


def generate_market_by_price_message(
    ticker: str,
    timestamp_iso_8601: str,
    gmt_offset: str,
    is_refresh: bool,
    message_number: int,
    summary_values: dict[str, list[str]] | None,
    map_entries: dict[tuple, dict[str, list[str]]] | None,
) -> str:
    message = f"{ticker},Market By Price,{timestamp_iso_8601},{gmt_offset},Raw,"
    if is_refresh:
        message += "REFRESH,,,,,3240,0,"
    else:
        message += "UPDATE,UNSPECIFIED,,,,3240,,"
    message += f"{message_number},0\n"
    message += generate_summary(summary_values)
    message += generate_map_entries(map_entries)
    return message


def generate_map_entries(map_entries: dict[tuple, dict[str, list[str]]] | None) -> str:
    """Generate a list of map entries

    Args:
        map_entries (dict[tuple, dict[str, list[str]]] | None): data for the generated map entries

    Returns:
        str: map entries built
    """
    message = ""
    if map_entries is not None:
        for key, value in map_entries.items():
            message += generate_map_entry(key[0], key[1], value)
    return message


def generate_map_entry(
    map_entry_type: str | None, map_entry_key: str | None, values: dict[str, list[str]] | None
) -> str:
    message = ""
    if values is not None:
        message += f",,,,MapEntry,,{map_entry_type},,,,,,{map_entry_key},{len(values.keys())}\n"
        if isinstance(values, dict):
            if len(values.keys()) > 0:
                for map_entry_key, value in values.items():
                    message += generate_field(
                        map_entry_key, value[0], value[1] if len(value) > 1 else ""
                    )
    return message


def generate_summary(
    values: dict[str, list[str]] | None,  # For example {"CURRENCY": ["999", "EUR"]} or None
) -> str:
    message = ""
    if isinstance(values, dict):
        message = f",,,,Summary,,,,,,,,,{len(values.keys())}\n"
        if len(values) > 0:
            for key, value in values.items():
                message += generate_field(key, value[0], value[1] if len(value) > 1 else "")
    return message


def generate_field(field_name: str, first_value: str, second_value: str = "") -> str:
    message = ""
    if field_name in known_fids:
        fid_code = known_fids[field_name][0]
        has_second_value = known_fids[field_name][1]
        if has_second_value and second_value == "":
            raise ValueError(
                f"The FID {field_name} needs to have two values but only one was present."
            )
        message += f",,,,FID,{fid_code},,{field_name},{first_value},{second_value}\n"
    else:
        raise ValueError("Unkown field name.")
    return message


@pytest.mark.parametrize("num_messages", [0, 100])
def test_find_market_by_price_lines_only_header(tmp_path: Path, num_messages: int) -> None:
    file = tmp_path / "test_file.csv"
    file.touch()
    test_messages = ""
    for num_message in range(num_messages):
        test_messages += generate_market_by_price_message(
            "TESTTICKER.MC",
            "2020-01-11T00:00:00.000000000Z",
            "+0",
            True,
            num_message,
            None,  # No summary line
            None,  # No map entries
        )
    file.write_text(test_messages)
    lines = find_market_by_price_lines(file)
    assert lines == [
        _ for _ in range(num_messages)
    ]  # Should find the headers in the correct places


@pytest.mark.parametrize("num_messages", [0, 100])
def test_find_market_by_price_lines_header_and_empty_summary(
    tmp_path: Path, num_messages: int
) -> None:
    file = tmp_path / "test_file.csv"
    file.touch()
    test_messages = ""
    for num_message in range(num_messages):
        test_messages += generate_market_by_price_message(
            "TESTTICKER.MC",
            "2020-01-11T00:00:00.000000000Z",
            "+0",
            True,
            num_message,
            {},  # Empty summary of no elements i.e. ",,,,Summary,,,,,,,,,0"
            None,  # No map entries
        )
    file.write_text(test_messages)
    lines = find_market_by_price_lines(file)
    assert lines == [
        _ for _ in range(0, num_messages * 2, 2)
    ]  # Should find the headers in the correct places


@pytest.mark.parametrize("num_messages", [0, 100])
def test_find_market_by_price_lines_header_and_summary(tmp_path: Path, num_messages: int) -> None:
    file = tmp_path / "test_file.csv"
    file.touch()
    test_messages = ""
    for num_message in range(num_messages):
        test_messages += generate_market_by_price_message(
            "TESTTICKER.MC",
            "2020-01-11T00:00:00.000000000Z",
            "+0",
            True,
            num_message,
            {"PROD_PERM": ["3240"], "DSPLY_NAME": ["TEST SA"], "CURRENCY": ["999", "TESTCOIN"]},
            None,
        )
    file.write_text(test_messages)
    lines = find_market_by_price_lines(file)

    # File contains a Market By Price header every 5 lines
    assert lines == [
        _ for _ in range(0, num_messages * 5, 5)
    ]  # Should find the headers in the correct places


@pytest.mark.parametrize("num_messages", [0, 1, 100])
@pytest.mark.parametrize("num_map_entries", [0, 1, 100])
def test_find_market_by_price_lines_header_and_summary_and_map_entries(
    tmp_path: Path, num_messages: int, num_map_entries: int
) -> None:
    file = tmp_path / "test_file.csv"
    file.touch()
    test_messages = ""
    for num_message in range(num_messages):
        random_map_entries = {}
        for index in range(num_map_entries):
            random_map_entries[("UPDATE", f"{index}.000000_B")] = {
                "BID_TIME": ["06:31:00.000000000"],
                "ORDER_PRC": ["100.0"],
                "ORDER_SIDE": ["1", "BID"],
                "NO_ORD": ["9"],
            }

        test_messages += generate_market_by_price_message(
            "TESTTICKER.MC",
            "2020-01-11T00:00:00.000000000Z",
            "+0",
            True,
            num_message,
            {"PROD_PERM": ["3240"], "DSPLY_NAME": ["TESTING SA"], "CURRENCY": ["999", "TESTCOIN"]},
            random_map_entries,
        )
    file.write_text(test_messages)
    lines = find_market_by_price_lines(file)

    assert lines == [
        _ for _ in range(0, num_messages * (5 + num_map_entries * 5), 5 + num_map_entries * 5)
    ]  # Should find the headers in the correct places


@pytest.mark.parametrize("num_map_entries", [0, 1, 100])
def test_flatten_map_entry_update_ok(num_map_entries):
    test_string = ""
    for _ in range(num_map_entries):
        test_string += generate_map_entry(
            "UPDATE",
            "1.000000_B",
            {
                "BID_TIME": ["06:31:00.000000000"],
                "ORDER_PRC": ["1.0"],
                "ORDER_SIDE": ["1", "BID"],
                "NO_ORD": ["9"],
                "ACC_SIZE": ["203"],
                "LV_TIM_MS": ["23460020"],
                "LV_DATE": ["2023-08-07"],
                "LV_TIM_NS": ["06:31:00.020570000"],
            },
        )
    result = flatten_map_entry(test_string)
    if num_map_entries == 0:
        assert result == [[]]
    elif num_map_entries > 0:
        expected = [
            [
                "ACC_SIZE",
                "BID_TIME",
                "LV_DATE",
                "LV_TIM_MS",
                "LV_TIM_NS",
                "MAP_ENTRY_KEY",
                "MAP_ENTRY_TYPE",
                "NO_ORD",
                "ORDER_PRC",
                "ORDER_SIDE",
            ]
        ]
        for _ in range(num_map_entries):
            expected.append(
                [
                    "203",
                    "06:31:00.000000000",
                    "2023-08-07",
                    "23460020",
                    "06:31:00.020570000",
                    "1.000000_B",
                    "UPDATE",
                    "9",
                    "1.0",
                    "BID",
                ]
            )
        assert result == expected


@pytest.mark.parametrize("num_map_entries", [0, 1, 100])
def test_flatten_map_entry_add_ok(num_map_entries):
    test_string = ""
    for _ in range(num_map_entries):
        test_string += generate_map_entry(
            "ADD",
            "1.000000_B",
            {
                "BID_TIME": ["06:31:00.000000000"],
                "ORDER_PRC": ["1.0"],
                "ORDER_SIDE": ["1", "BID"],
                "NO_ORD": ["9"],
                "ACC_SIZE": ["203"],
                "LV_TIM_MS": ["23460020"],
                "LV_DATE": ["2023-08-07"],
                "LV_TIM_NS": ["06:31:00.020570000"],
            },
        )
    result = flatten_map_entry(test_string)
    if num_map_entries == 0:
        assert result == [[]]
    elif num_map_entries > 0:
        expected = [
            [
                "ACC_SIZE",
                "BID_TIME",
                "LV_DATE",
                "LV_TIM_MS",
                "LV_TIM_NS",
                "MAP_ENTRY_KEY",
                "MAP_ENTRY_TYPE",
                "NO_ORD",
                "ORDER_PRC",
                "ORDER_SIDE",
            ]
        ]
        for _ in range(num_map_entries):
            expected.append(
                [
                    "203",
                    "06:31:00.000000000",
                    "2023-08-07",
                    "23460020",
                    "06:31:00.020570000",
                    "1.000000_B",
                    "ADD",
                    "9",
                    "1.0",
                    "BID",
                ]
            )
        assert result == expected


@pytest.mark.parametrize("num_map_entries", [0, 1, 100])
def test_flatten_map_entry_delete_ok(num_map_entries):
    test_string = ""
    for _ in range(num_map_entries):
        test_string += generate_map_entry("DELETE", "1.000000_B", {})
    result = flatten_map_entry(test_string)
    if num_map_entries == 0:
        assert result == [[]]
    elif num_map_entries > 0:
        expected = [
            [
                "MAP_ENTRY_KEY",
                "MAP_ENTRY_TYPE",
            ]
        ]
        for _ in range(num_map_entries):
            expected.append(["1.000000_B", "DELETE"])
        assert result == expected


@pytest.mark.parametrize("num_messages", [0, 1, 100, 500])  # Number of Market By Price messages
@pytest.mark.parametrize(
    "num_map_entries", [0, 1, 50]
)  # Number of MapEntry messages per Market By Price message
def test_run_ok(tmp_path: Path, num_messages: int, num_map_entries: int) -> None:
    file = tmp_path / "test_file.csv"
    file.touch()
    # Build expected DataFrame
    expected = pl.DataFrame()
    test_messages = ""
    for num_message in range(num_messages):
        random_map_entries = {}
        market_by_price_df = pl.DataFrame()
        for index in range(num_map_entries):
            random_map_entries[("UPDATE", f"{index}.000000_B")] = {
                "BID_TIME": ["06:31:00.000000000"],
                "ORDER_PRC": ["100.0"],
                "ORDER_SIDE": ["1", "BID"],
                "NO_ORD": ["9"],
            }

        for key, map_entry in random_map_entries.items():
            row = pl.DataFrame({k: v[-1] for k, v in map_entry.items()})
            row = row.with_columns(pl.lit(key[0]).alias("MAP_ENTRY_TYPE"))
            row = row.with_columns(pl.lit(key[1]).alias("MAP_ENTRY_KEY"))
            market_by_price_df = pl.concat([market_by_price_df, row], how="vertical")

        test_messages += generate_market_by_price_message(
            "TESTTICKER.MC",
            "2020-01-11T00:00:00.000000000Z",
            "+0",
            True,
            num_message,
            {"PROD_PERM": ["3240"], "DSPLY_NAME": ["TESTING SA"], "CURRENCY": ["999", "TESTCOIN"]},
            random_map_entries,
        )
        market_by_price_df = market_by_price_df.with_columns(
            pl.lit("REFRESH").alias("MARKET_MESSAGE_TYPE")
        )
        market_by_price_df = market_by_price_df.with_columns(
            pl.lit("TESTTICKER.MC").alias("TICKER")
        )
        market_by_price_df = market_by_price_df.with_columns(
            pl.lit("2020-01-11T00:00:00.000000000Z").alias("TIMESTAMP")
        )
        market_by_price_df = market_by_price_df.with_columns(pl.lit("+0").alias("GMT_OFFSET"))
        market_by_price_df = market_by_price_df.with_columns(pl.lit("3240").alias("PROD_PERM"))
        market_by_price_df = market_by_price_df.with_columns(
            pl.lit("TESTING SA").alias("DSPLY_NAME")
        )
        market_by_price_df = market_by_price_df.with_columns(pl.lit("TESTCOIN").alias("CURRENCY"))

        expected = pl.concat([expected, market_by_price_df.clone()])

    # Force column order in expected order. I.e. the final fill
    for col in supplement:
        if col not in expected.columns:
            expected = expected.with_columns(pl.lit("").alias(col))
    expected = expected.select(sorted(expected.columns))

    file.write_text(test_messages)

    assert run(Path(file), Path(tmp_path))

    if test_messages != "":
        res = pl.read_parquet(tmp_path / "part-*.parquet")
        assert_frame_equal(res, expected)
    else:
        # If the message was empty, no parquet file should be generated
        assert not all([file.suffix.endswith("parquet") for file in Path(tmp_path).iterdir()])


@pytest.mark.parametrize(
    "num_messages", [0, 1]
)  # Number of Market By Price messages. IMPORTANT: flatten_market_by_price expects one Market By Price message
@pytest.mark.parametrize(
    "num_map_entries", [0, 1, 100]
)  # Number of MapEntry messages per Market By Price message
def test_flatten_market_by_price_ok(
    tmp_path: Path, num_messages: int, num_map_entries: int
) -> None:
    file = tmp_path / "test_file.csv"
    file.touch()
    # Build expected DataFrame
    expected = pl.DataFrame()
    test_messages = ""
    for num_message in range(num_messages):
        random_map_entries = {}
        market_by_price_df = pl.DataFrame()
        for index in range(num_map_entries):
            random_map_entries[("UPDATE", f"{index}.000000_B")] = {
                "BID_TIME": ["06:31:00.000000000"],
                "ORDER_PRC": ["100.0"],
                "ORDER_SIDE": ["1", "BID"],
                "NO_ORD": ["9"],
            }

        for key, map_entry in random_map_entries.items():
            row = pl.DataFrame({k: v[-1] for k, v in map_entry.items()})
            row = row.with_columns(pl.lit(key[0]).alias("MAP_ENTRY_TYPE"))
            row = row.with_columns(pl.lit(key[1]).alias("MAP_ENTRY_KEY"))
            market_by_price_df = pl.concat([market_by_price_df, row], how="vertical")

        test_messages += generate_market_by_price_message(
            "TESTTICKER.MC",
            "2020-01-11T00:00:00.000000000Z",
            "+0",
            True,
            num_message,
            {"PROD_PERM": ["3240"], "DSPLY_NAME": ["TESTING SA"], "CURRENCY": ["999", "TESTCOIN"]},
            random_map_entries,
        )
        market_by_price_df = market_by_price_df.with_columns(
            pl.lit("REFRESH").alias("MARKET_MESSAGE_TYPE")
        )
        market_by_price_df = market_by_price_df.with_columns(
            pl.lit("TESTTICKER.MC").alias("TICKER")
        )
        market_by_price_df = market_by_price_df.with_columns(
            pl.lit("2020-01-11T00:00:00.000000000Z").alias("TIMESTAMP")
        )
        market_by_price_df = market_by_price_df.with_columns(pl.lit("+0").alias("GMT_OFFSET"))
        market_by_price_df = market_by_price_df.with_columns(pl.lit("3240").alias("PROD_PERM"))
        market_by_price_df = market_by_price_df.with_columns(
            pl.lit("TESTING SA").alias("DSPLY_NAME")
        )
        market_by_price_df = market_by_price_df.with_columns(pl.lit("TESTCOIN").alias("CURRENCY"))

        expected = pl.concat([expected, market_by_price_df.clone()])

    res = flatten_market_by_price(test_messages)

    assert_frame_equal(res, expected, check_column_order=False)
